#!/usr/bin/env python3

from stepwise import Quantity
from pathlib import Path
from voluptuous import Schema, Optional, Required, All, Any, Coerce
from mako.template import Template
from ..model import REAGENT_CLASSES, Database, Plasmid, find
from ..utils import check_tag, parse_bool
from ..fields import parse_fields, parse_fields_list
from ..errors import LoadError

schema = Schema({
        'format': 'excel',
        Required('reagent'): Any(*REAGENT_CLASSES),
        Required('path'): All(str, Coerce(Path)),
        Optional('worksheet'): str,
        Optional('sequence'): str,
        Optional('tag_template'): str,
        'columns': {str: Any(
            'tag', 'seq', 'molecule', 'synthesis', 'cleanups', 'ready', 'name',
            'alt_names', 'date', 'desc', 'length', 'conc', 'mw', 'circular',
            'resistance', 'antibiotics', 'origin', 'parent_strain', 'plasmids',
        )},
})
default_config = {
        'columns': {
            'Tag': 'tag',
            'Sequence': 'seq',
            'Molecule': 'molecule',
            'Synthesis': 'synthesis',
            'Cleanups': 'cleanups',
            'Ready': 'ready',
            'Name': 'name',
            'Cross-refs': 'alt_names',
            'Date': 'date',
            'Description': 'desc',
            'Length': 'length',
            'Conc': 'conc',
            'MW': 'mw',
            'Circular': 'circular',
            'Resistance': 'resistance',
            'Antibiotics': 'antibiotics',
            'Origin': 'origin',
            'Strain': 'parent_strain',
            'Plasmids': 'plasmids',
        }
}

def load(db, config):
    import pandas as pd
    from warnings import catch_warnings, filterwarnings

    path = config['path']
    sheet = config.get('worksheet', 0)
    reagent = config['reagent']
    columns = config['columns']
    columns_user = {v: k for k, v in columns.items()}
    
    seq_path_template = config.get('sequence')
    reagent_cls = REAGENT_CLASSES[reagent]

    if not path.exists():
        raise LoadError("database not found: {path}", path=path)

    with catch_warnings():
        filterwarnings(
                'ignore',
                category=UserWarning,
                message="Workbook contains no default style, apply openpyxl's default",
        )
        filterwarnings(
                'ignore',
                category=DeprecationWarning,
                message="`np.float` is a deprecated alias for the builtin `float`.",
        )
        df = pd.read_excel(path, sheet_name=sheet)

    df = df.rename(columns=columns)
    df = df.astype(object).where(pd.notnull(df), None)

    def _find_tag(kwargs, config):
        try:
            return kwargs.pop('tag')
        except KeyError:
            pass

        try:
            tag_template = config['tag_template']
        except KeyError:
            err = LoadError(tag_col=columns_user.get('tag', 'tag'))
            err.brief = "reagent must have tag"
            err.blame += "expected to find tag in {tag_col!r} column"
            raise err from None

        try:
            return Template(tag_template).render(i=i, **row)
        except Exception as err1:
            err2 = LoadError(tag_template=tag_template, err=err1)
            err2.brief = "can't create tag from template"
            err2.info += "template: {tag_template}"
            err2.blame += "{err.__class__.__name__}: {err}"
            raise err2 from None

    for i, row in df.iterrows():
        kwargs = {k: v for k, v in row.items() if v is not None}

        with LoadError.add_info('path: {path}', 'row: {i}', path=path, row=row, i=i):
            tag = _find_tag(kwargs, config)

        if not kwargs.get('seq') and hasattr(reagent_cls, 'seq'):
            kwargs['seq'] = _defer(
                    _load_seq_from_tag, db, tag, seq_path_template)
        if x := kwargs.get('alt_names'):
            kwargs['alt_names'] = _comma_list(x)
        if x := kwargs.get('conc'):
            kwargs['conc'] = _defer(Quantity.from_string, x)
        if x := kwargs.get('synthesis'):
            kwargs['synthesis'] = _defer(parse_fields, x)
        if x := kwargs.get('cleanups'):
            kwargs['cleanups'] = _defer(parse_fields_list, x)
        if x := kwargs.get('ready'):
            kwargs['ready'] = _defer(parse_bool, x)
        if x := kwargs.get('circular'):
            kwargs['circular'] = _defer(parse_bool, x)
        if x := kwargs.get('resistance'):
            kwargs['resistance'] = _comma_list(x)
        if x := kwargs.get('antibiotics'):
            kwargs['antibiotics'] = _comma_list(x)
        if x := kwargs.get('plasmids'):
            kwargs['plasmids'] = _defer(
                    find, db, _comma_list(x), reagent_cls=Plasmid)

        db[tag] = reagent_cls(**kwargs)

def _defer(f, *args, **kwargs):
    """
    Create a closure that calls the given function with the given arguments.

    The point of this function is to avoid the surprising behavior that can 
    occur if you define a closure in a scope where variables are changing (e.g. 
    in a for-loop).  The confusing thing is that closures have access to the 
    scope they were defined in, but only as it exists when they are ultimately 
    called.  So if the scope changes between when the closure is defined and 
    when it's called, the closure will use the final value of any variables.

    This function serves to create a static local scope containing the 
    variables needed by the closure, which avoids the problem.

    More information: 
    https://stackoverflow.com/questions/10452770/python-lambdas-binding-to-local-values
    """
    class defer:

        def __repr__(self):
            return '<deferred>'

        def __call__(self):
            return f(*args, **kwargs)

    return defer()

def _load_seq_from_tag(db, tag, path_template):
    path = _path_from_tag(db, tag, path_template)
    return _load_seq(path)

def _load_seq(path):
    import autosnapgene as snap

    if not path.exists():
        return

    if path.suffix == '.dna':
        return snap.parse(path).dna_sequence
    if path.suffix == '.prot':
        return snap.parse(path).protein_sequence

    err = LoadError(path=path)
    err.brief = "can't load sequence info from {path.suffix!r} files"
    err.info += "path: {path}"
    raise err

def _path_from_tag(db, tag, path_template):
    path_str = path_template.format(
            tag=tag,
            tag_match=_TagMatch(db, tag),
    )
    return Path(path_str)

def _comma_list(list_str):
    return [x.strip() for x in list_str.split(',')]

class _TagMatch:

    def __init__(self, db, tag):
        self.db = db
        self.tag = tag
        self.match = None

    def __getitem__(self, key):
        # Defer matching the regular expression until we're asked for it, so 
        # that we can raise a useful error message if it doesn't match.
        if not self.match:
            self.match = check_tag(self.db, self.tag)

        return self.match.group(key)




#!/usr/bin/env python3

from freezerbox.config import *
from freezerbox.utils import cd
from param_helpers import *

@parametrize_from_file(
        schema=Schema({
            'files': {str: str},
            Optional('cwd', default='.'): str,
            'expected_config': dict,
            'expected_paths': dict,
        }),
        indirect=['files'],
)
def test_load_config(files, cwd, expected_config, expected_paths):
    load_config.cache_clear()

    with_paths = Namespace(
            DIR=files,
            BUILTIN_CONF=BUILTIN_CONF,
    )
    expected_paths = with_paths.eval(expected_paths)

    with cd(files / cwd):
        config = load_config()

        # Remove feature sequences, just to simplify test files:
        for feat in config.get('features', []):
            feat.pop('seq', None)

        assert config == expected_config
        assert config.paths == expected_paths

def test_dict_view():
    d = {'a': 'b'}
    v = DictView(d)

    assert v == d
    assert v == v
    assert len(v) == 1
    assert repr(v) == "DictView({'a': 'b'})"
    assert list(iter(v)) == ['a']
    assert v['a'] == 'b'




#!/usr/bin/env python3

import freezerbox, pytest
import parametrize_from_file

from byoc.errors import Log
from more_itertools import one, zip_equal
from re_assert import Matches
from mock_model import mock_plugins, MockMolecule
from operator import attrgetter
from pprint import pprint
from param_helpers import *

class MockObj:
    pass

@parametrize_from_file(
        schema=Schema({
            'db': dict,
            'obj': str,
            Optional('config_cls', default='class MockConfig(ReagentConfig): pass'): str,
            Optional('db_access', default='cache'): str,
            'key': with_py.eval,
            'expected': [with_py.eval],
            'info': [str],
    }),
)
def test_reagent_config(db, config_cls, obj, db_access, key, expected, info, monkeypatch, mock_plugins):
    with_mock_obj = Namespace(
            MockObj=MockObj,
            LoadError=freezerbox.LoadError,
    )
    with_reagent_config = Namespace(
        ReagentConfig=freezerbox.ReagentConfig,
        MockMolecule=MockMolecule,
        attrgetter=attrgetter,
    )

    db = eval_db(db)
    obj = with_mock_obj.exec(obj)['obj']
    config_cls = with_reagent_config.exec(config_cls)['MockConfig']
    config = config_cls(obj)
    layer = one(config.load())

    if db_access == 'none':
        pass

    if db_access == 'cache':
        layer.db = db

    if db_access.startswith('obj'):
        attr = db_access.split('.')[1] if '.' in db_access else 'db'
        setattr(obj, attr, db)

    if db_access == 'load':
        monkeypatch.setattr(freezerbox.model, 'load_db', lambda: db)
    elif db_access == 'load-err':
        def fail_to_load(): raise freezerbox.LoadError("mock error")
        monkeypatch.setattr(freezerbox.model, 'load_db', fail_to_load)
    else:
        monkeypatch.setattr(freezerbox.model, 'load_db', lambda: NotImplemented)

    log = Log()
    values = list(layer.iter_values(key, log))

    try:
        assert values == expected
    finally:
        print(log.format())

    for actual, pattern in zip_equal(log.message_strs, info):
        Matches(pattern).assert_matches(actual)

@parametrize_from_file(
        schema=Schema({
            'db': dict,
            'config_cls': str,
            'products': {str: Coerce(int)},
            Optional('products_attr', default='products'): str,
            'key': with_py.eval,
            **with_freeze.error_or({
                'expected': [with_freeze.eval],
                'info': [str],
            }),
    }),
)
def test_product_configs(db, config_cls, products, products_attr, key, expected, info, error, mock_plugins):
    with_configs = Namespace(
            ProductConfig=freezerbox.ProductConfig,
            MakerConfig=freezerbox.MakerConfig,
            PrecursorConfig=freezerbox.PrecursorConfig,
            MockMolecule=MockMolecule,
            attrgetter=attrgetter,
    )

    db = eval_db(db)
    config_cls = with_configs.exec(config_cls)['MockConfig']
    obj = MockObj()

    if products_attr:
        setattr(obj, products_attr, [
            db[k].make_intermediate(i)
            for k,i in products.items()
        ])

    config = config_cls(obj)
    layer = one(config.load())

    with error:
        log = Log()
        values = list(layer.iter_values(key, log))
        pprint(log.message_strs)

        assert values == expected

        for actual, pattern in zip_equal(log.message_strs, info):
            Matches(pattern).assert_matches(actual)




#!/usr/bin/env python3

import re, os
from stepwise import Quantity
from more_itertools import split_when, zip_equal, first
from contextlib import contextmanager
from inform import did_you_mean
from .errors import ParseError, only_raise

NO_DEFAULT = object()

TIME_CONVERSION_FACTORS = {
    's':        60*60,
    'sec':      60*60,
    'second':   60*60,
    'seconds':  60*60,
    'm':        60,
    'min':      60,
    'minute':   60,
    'minutes':  60,
    'h':        1,
    'hr':       1,
    'hour':     1,
    'hours':    1,
}
TEMP_CONVERSION_FACTORS = {
        '°C': 1,
         'C': 1,
}
VOLUME_CONVERSION_FACTORS = {
        'nL': 1e9,
        'uL': 1e6,
        'µL': 1e6,
        'mL': 1e3,
         'L': 1,
}
MASS_CONVERSION_FACTORS = {
        'pg': 1e12,
        'ng': 1e9,
        'ug': 1e6,
        'µg': 1e6,
        'mg': 1e3,
         'g': 1,
        'kg': 1e-3,
}
CONC_CONVERSION_FACTORS_MOLARITY = {
        'pM': 1e12,
        'nM': 1e9,
        'uM': 1e6,
        'µM': 1e6,
        'mM': 1e3,
         'M': 1,
}
CONC_CONVERSION_FACTORS_DENSITY = {
        'ng/uL': 1e3,
        'ng/µL': 1e3,
        'µg/mL': 1e3,
        'ug/mL': 1e3,
        'µg/µL': 1,
        'ug/uL': 1,
        'mg/mL': 1,
}
CONC_UNITS = {
        *CONC_CONVERSION_FACTORS_MOLARITY,
        *CONC_CONVERSION_FACTORS_DENSITY,
}
SIZE_CONVERSION_FACTORS = {
        'bp': 1e3,
        'kb': 1,
}

def normalize_seq(raw_seq):
    # Ignore nonstandard nucleotides; they're too hard to deal with properly.

    # This regular expression works because `re.sub()` only substitutes the 
    # left-most occurrence of any overlapping patterns.  The non-greedy * is 
    # necessary to avoid eliding everything between the first and last 
    # nonstandard nucleotide.
    seq = re.sub(r'/.*?/', 'X', raw_seq.strip())

    return seq.upper()

def reverse_complement(seq):
    from Bio.Seq import Seq
    return str(Seq(seq).reverse_complement())

def calc_sequence_identity(
        seq, ref_seq, *,
        match_score=1,
        mismatch_score=-1,
        open_gap_score=-0.5,
        extend_gap_score=-0.5,
):
    """
    Return the fraction of nucleotides in the given sequence that can be 
    perfectly aligned to the reference sequence.

    The basic algorithm is: perform a local sequence alignment, then count how 
    many positions in the input sequence are present in the alignment and 
    identical to the reference.  There are other ways to calculate percent 
    identity (e.g. different choice of denominator, different treatment of 
    gaps), but this should be reasonable.
    """
    from Bio.Align import PairwiseAligner

    if not seq:
        return 0

    seq = seq.upper()
    ref_seq = ref_seq.upper()

    aligner = PairwiseAligner()
    aligner.mode = 'local'
    aligner.match_score = match_score
    aligner.mismatch_score = mismatch_score
    aligner.open_gap_score = open_gap_score
    aligner.extend_gap_score = extend_gap_score

    results = aligner.align(seq, ref_seq)

    try:
        result = first(results)
    except ValueError:
        return 0

    aligned_chars = []

    for (i,j), (k,l) in zip(*result.aligned):
        aligned_chars += zip_equal(seq[i:j], ref_seq[k:l])

    n_identities = sum(a == b for a,b in aligned_chars)
    n_total = len(seq)

    return n_identities / n_total

def calc_sequence_identity_with_rc(seq, ref_seq, **kwargs):
    ref_seqs = [ref_seq, reverse_complement(ref_seq)]
    scores = [
            calc_sequence_identity(seq, x, **kwargs)
            for x in ref_seqs
    ]
    return max(scores)

@only_raise(ParseError)
def check_tag(db, tag):
    tag_pattern = db.config.get('tag_pattern', '.*')

    if m := re.fullmatch(tag_pattern, tag):
        return m

    err = ParseError(db=db, tag=tag, pattern=tag_pattern)
    err.brief = "tag doesn't match expected pattern"
    err.info += "relevant config: {db.config.paths[tag_pattern]}"
    err.blame += "expected {tag!r} to match {pattern!r}"
    raise err

@only_raise(ParseError)
def parse_bool(bool_str):
    if bool_str.lower() in ('1', '+', 'x', 'y', 'yes', 'true'):
        return True
    if bool_str.lower() in ('0', '-',  '', 'n', 'no', 'false'):
        return False

    raise ParseError(f"can't interpret {bool_str!r} as a bool")

@only_raise(ParseError)
def parse_time(time_str, default_unit=None):

    # First, try parsing the time using a custom syntax, e.g. '1m30':
    time_pattern = r'(\d+)([hm])(\d+)'
    if m := re.fullmatch(time_pattern, time_str):
        value = 60 * int(m.group(1)) + int(m.group(3))
        unit = {'h': 'm', 'm': 's'}[m.group(2)]
        return Quantity(value, unit)

    # If that doesn't work, try parsing the time normally, e.g. '1 min'
    return _parse_quantity(
            time_str, TIME_CONVERSION_FACTORS, default_unit, 'time')

@only_raise(ParseError)
def parse_time_s(time_str, default_unit=None):
    time = parse_time(time_str, default_unit)
    time = time.convert_unit('s', TIME_CONVERSION_FACTORS)
    return time.value

@only_raise(ParseError)
def parse_time_m(time_str, default_unit=None):
    time = parse_time(time_str, default_unit)
    time = time.convert_unit('m', TIME_CONVERSION_FACTORS)
    return time.value

@only_raise(ParseError)
def parse_time_h(time_str, default_unit=None):
    time = parse_time(time_str, default_unit)
    time = time.convert_unit('h', TIME_CONVERSION_FACTORS)
    return time.value

def format_time_s(x):
    if x < 60:
        return f'{x}s'

    min = x // 60
    sec = x % 60

    return f'{min}m{f"{sec:02}" if sec else ""}'

def format_time_m(x):
    if x < 60:
        return f'{x}m'

    hr = x // 60
    min = x % 60

    return f'{hr}h{f"{min:02}" if min else ""}'

@only_raise(ParseError)
def parse_temp(temp_str, default_unit=None):
    return _parse_quantity(
            temp_str, TEMP_CONVERSION_FACTORS, default_unit, 'temperature')

@only_raise(ParseError)
def parse_temp_C(temp_str, default_unit=None):
    return _parse_and_convert_quantity(
            temp_str, TEMP_CONVERSION_FACTORS, default_unit, '°C', 'temperature')

@only_raise(ParseError)
def parse_volume(vol_str, default_unit=None):
    return _parse_quantity(
            vol_str, VOLUME_CONVERSION_FACTORS, default_unit, 'volume')

@only_raise(ParseError)
def parse_volume_uL(vol_str, default_unit=None):
    return _parse_and_convert_quantity(
            vol_str, VOLUME_CONVERSION_FACTORS, default_unit, 'µL', 'volume')

@only_raise(ParseError)
def parse_volume_mL(vol_str, default_unit=None):
    return _parse_and_convert_quantity(
            vol_str, VOLUME_CONVERSION_FACTORS, default_unit, 'mL', 'volume')

@only_raise(ParseError)
def parse_mass(mass_str, default_unit=None):
    return _parse_quantity(
            mass_str, MASS_CONVERSION_FACTORS, default_unit, 'mass')

@only_raise(ParseError)
def parse_mass_ug(mass_str, default_unit=None):
    return _parse_and_convert_quantity(
            mass_str, MASS_CONVERSION_FACTORS, default_unit, 'µg', 'mass')

@only_raise(ParseError)
def parse_mass_mg(mass_str, default_unit=None):
    return _parse_and_convert_quantity(
            mass_str, MASS_CONVERSION_FACTORS, default_unit, 'mg', 'mass')

@only_raise(ParseError)
def parse_conc(conc_str, default_unit=None):
    return _parse_quantity(conc_str, CONC_UNITS, default_unit, 'concentration')

@only_raise(ParseError)
def parse_conc_nM(conc_str, mw=None, default_unit=None):
    conc = parse_conc(conc_str, default_unit)
    conc = convert_conc_unit(conc, mw, 'nM')
    return conc.value

@only_raise(ParseError)
def parse_conc_uM(conc_str, mw=None, default_unit=None):
    conc = parse_conc(conc_str, default_unit)
    conc = convert_conc_unit(conc, mw, 'µM')
    return conc.value

@only_raise(ParseError)
def parse_conc_ng_uL(conc_str, mw=None, default_unit=None):
    conc = parse_conc(conc_str, default_unit)
    conc = convert_conc_unit(conc, mw, 'ng/µL')
    return conc.value

@only_raise(ParseError)
def parse_conc_ug_uL(conc_str, mw=None, default_unit=None):
    conc = parse_conc(conc_str, default_unit)
    conc = convert_conc_unit(conc, mw, 'µg/µL')
    return conc.value

@only_raise(ParseError)
def convert_conc_unit(conc, mw, new_unit):
    def pick_conversion_factors():
        if mw is not None:
            return {
                    **CONC_CONVERSION_FACTORS_MOLARITY,
                    **{
                        k: mw * v
                        for k, v in CONC_CONVERSION_FACTORS_DENSITY.items()
                    },
            }

        if conc.unit in CONC_CONVERSION_FACTORS_MOLARITY:
            return CONC_CONVERSION_FACTORS_MOLARITY

        if conc.unit in CONC_CONVERSION_FACTORS_DENSITY:
            return CONC_CONVERSION_FACTORS_DENSITY

        return {}

    return conc.convert_unit(new_unit, pick_conversion_factors())

@only_raise(ParseError)
def parse_size(size_str, default_unit=None):
    return _parse_quantity(size_str, SIZE_CONVERSION_FACTORS, default_unit, 'size')

@only_raise(ParseError)
def parse_size_bp(size_str, default_unit=None):
    return _parse_and_convert_quantity(
            size_str, SIZE_CONVERSION_FACTORS, default_unit, 'bp', 'size')

@only_raise(ParseError)
def parse_size_kb(size_str, default_unit=None):
    return _parse_and_convert_quantity(
            size_str, SIZE_CONVERSION_FACTORS, default_unit, 'kb', 'size')

def parse_stranded_molecule(molecule, default_strandedness=None):
    """
    Parse the molecule type (e.g. suitable input for Biopython functions) 
    and strandedness from a molecule string.  
    """
    if isinstance(molecule, tuple):
        return molecule

    stranded_molecules = {
            'dna':   (0, 'DNA'),
            'ssdna': (1, 'DNA'),
            'dsdna': (2, 'DNA'),
            'rna':   (0, 'RNA'),
            'ssrna': (1, 'RNA'),
            'dsrna': (2, 'RNA'),
    }
    try:
        strandedness, molecule = stranded_molecules[molecule.lower()]
    except KeyError:
        err = ParseError(molecule=molecule)
        err.brief = "unknown molecule type: {molecule!r}"
        err.info += f"expected: {', '.join(map(repr, stranded_molecules))}"
        raise err

    if not strandedness:
        strandedness = default_strandedness

    if not strandedness:
        strandedness = {'DNA': 2, 'RNA': 1}[molecule]

    return molecule, strandedness

def mw_from_length(len, molecule='dna'):
    molecule, strandedness = parse_stranded_molecule(molecule)

    # Parameters from: 
    # https://www.thermofisher.com/us/en/home/references/ambion-tech-support/rna-tools-and-calculators/dna-and-rna-molecular-weights-and-conversions.html
    params = {
            'DNA': (303.7, 79),
            'RNA': (320.5, 159),
    }
    m, b = params[molecule]
    return (m * len + b) * strandedness

def unanimous(
        items,
        default=NO_DEFAULT,
        err_empty=ValueError("empty iterable"),
        err_multiple=lambda v1, v2: ValueError(f"found multiple values: {v1!r}, {v2!r}"),
    ):
    it = iter(items)

    try:
        value = next(it)
    except StopIteration:
        if default is not NO_DEFAULT:
            return default
        else:
            raise err_empty

    for next_value in it:
        if next_value != value:
            raise err_multiple(value, next_value)

    return value

def join_lists(x):
    from itertools import chain
    return list(chain(*x))

def join_dicts(x):
    # Originally I used `dict(ChainMap(*x))`, but I reimplemented it using 
    # basic for-loops to maintain insertion order.
    out = {}
    for d in x:
        for k, v in d.items():
            if k not in out:
                out[k] = v
    return out

def join_sets(x):
    return set.union(set(), *x)

@contextmanager
def cd(dir):
    try:
        prev_cwd = os.getcwd()
        os.chdir(dir)
        yield

    finally:
        os.chdir(prev_cwd)


def _parse_quantity(quantity_str, expected_units, default_unit, unit_type):
    try:
        if default_unit:
            quantity = Quantity.from_string_or_float(quantity_str, default_unit)
        else:
            quantity = Quantity.from_string(quantity_str)

    except ValueError:
        raise ParseError(
                "can't interpret {quantity_str!r} as a {unit_type}",
                quantity_str=quantity_str,
                unit_type=unit_type,
        ) from None

    if quantity.unit not in expected_units:
        raise ParseError(
                lambda e: f"can't interpret {e.quantity_str!r} as a {unit_type}, did you mean '{Quantity(e.quantity.value, did_you_mean(e.quantity.unit, e.expected_units))}'?",
                quantity=quantity,
                quantity_str=quantity_str,
                expected_units=expected_units,
                unit_type=unit_type,
        )

    return quantity

def _parse_and_convert_quantity(quantity_str, conversion_factors, default_unit, desired_unit, unit_type):
    quantity = _parse_quantity(
            quantity_str, conversion_factors, default_unit, unit_type)
    converted = quantity.convert_unit(desired_unit, conversion_factors)
    return converted.value





#!/usr/bin/env python3

import freezerbox
import pytest
import parametrize_from_file

from parametrize_from_file import star
from parametrize_from_file.voluptuous import Namespace
from voluptuous import Schema, Invalid, Coerce, And, Or, Optional
from pathlib import Path

TEST_DIR = Path(__file__).parent

def eval_db(reagents):
    schema = Schema({
            str: with_freeze.eval,
    })
    reagents = schema(reagents)

    meta = reagents.pop('meta', {})
    config = meta.get('config', {})
    paths = meta.get('paths', {})

    config = freezerbox.Config(config, paths)
    db = freezerbox.Database(config)

    for tag, reagent in reagents.items():
        db[tag] = reagent

    if not db.name:
        db.name = 'mock'

    return db

def empty_ok(x):
    return Or(x, And('', lambda y: type(x)()))

def approx_Q(x):
    from stepwise import Quantity
    q = Quantity.from_anything(x)
    return pytest.approx(q, abs=Quantity(1e-6, q.unit))

with_py = Namespace(
        'from operator import attrgetter',
)
with_nx = Namespace(
        'import networkx as nx',
)
with_pytest = Namespace(
        'from pytest import *',
)
with_sw = Namespace(
        with_py,
        with_pytest,
        'from stepwise import Quantity, Q, pl, ul, ol, table',
        approx_Q=approx_Q,
)
with_freeze = Namespace(
        with_sw,
        'import freezerbox; from freezerbox import *',
        'from mock_model import *',
        TEST_DIR=TEST_DIR,
)

@pytest.fixture
def files(request, tmp_path):
    for name, contents in request.param.items():
        p = tmp_path / name
        p.parent.mkdir(parents=True, exist_ok=True)
        p.write_text(contents)

    return tmp_path




#!/usr/bin/env python3

from param_helpers import *

class DummyItem:

    def __init__(self, **kwargs):
        self.kwargs = kwargs

    def __repr__(self):
        kwargs = ', '.join(f'{k}={v!r}' for k,v in self.kwargs.items())
        return f'DummyItem({kwargs})'

    def __eq__(self, other):
        return self.kwargs == other.kwargs

    def __getattr__(self, name):
        return self.kwargs[name]

with_dummy = Namespace(DummyItem=DummyItem)

@parametrize_from_file(
        schema=Schema({
            'grouper': with_freeze.eval,
            'items': with_py.eval,
            Optional('key', default='lambda x: x'): with_py.eval,
            'expected': empty_ok([{
                'value': with_pytest.eval,
                'items': with_pytest.eval,
            }]),
        }),
)
def test_group_by(grouper, items, key, expected):
    actual = [
            (k, list(it))
            for k, it in grouper(items, key=key)
    ]
    expected = [
            (x['value'], list(x['items']))
            for x in expected
    ]
    assert actual == expected

@parametrize_from_file(
        schema=Schema({
            'items': empty_ok([with_dummy.eval]),
            'group_by': empty_ok({str: with_freeze.eval}),
            'merge_by': empty_ok({str: with_freeze.eval}),
            Optional('keys', default={}): empty_ok({str: with_freeze.eval}),
            'expected': empty_ok([{
                'attrs': empty_ok({str: with_pytest.eval}),
                'items': [with_dummy.eval],
            }]),
        }),
)
def test_iter_combos(items, group_by, merge_by, keys, expected):
    actual = [
            (attrs, list(items))
            for attrs, items in freezerbox.iter_combos(
                items,
                group_by=group_by,
                merge_by=merge_by,
                keys=keys,
            )
    ]
    expected = [
            (x['attrs'], list(x['items']))
            for x in expected
    ]
    assert actual == expected



#!/usr/bin/env python3

import rtoml

from pathlib import Path
from functools import lru_cache
from mergedeep import merge
from collections.abc import Mapping

BUILTIN_CONF = Path(__file__).parent / 'conf.toml'

@lru_cache
def load_config():
    config = rtoml.load(BUILTIN_CONF)
    config_paths = _fill(config, BUILTIN_CONF)

    cwd = Path.cwd()
    dirs = [cwd, *cwd.parents]

    for dir in reversed(dirs):
        paths = [
                dir / '.config' / 'freezerbox' / 'conf.toml',
                dir / '.freezerboxrc',
        ]
        for path in paths:
            if path.exists():
                local_config = rtoml.load(path)
                local_config_paths = _fill(local_config, path)

                merge(config, local_config)
                merge(config_paths, local_config_paths)

    return Config(config, config_paths)

class DictView(Mapping):

    def __init__(self, dict):
        self.dict = dict

    def __repr__(self):
        return f'{self.__class__.__name__}({self.dict!r})'

    def __iter__(self):
        return self.dict.__iter__()

    def __len__(self):
        return self.dict.__len__()

    def __getitem__(self, key):
        return self.dict.__getitem__(key)

class Config(DictView):

    def __init__(self, config, paths={}):
        super().__init__(config)

        # Guarantee that there is a path for every config option.  This should 
        # never come up if loading a database via `load_db()`, but it does make 
        # it easier to construct databases on the fly.  This happens a lot in 
        # testing.
        all_paths = _fill(config, 'unspecified')
        merge(all_paths, paths)

        self.paths = DictView(all_paths)

def _fill(given, value):
    return {
            k: _fill(v, value) if isinstance(v, dict) else value
            for k, v in given.items()
    }



#!/usr/bin/env python3

from freezerbox import load_db
from param_helpers import *

# Only cases where the database cannot be loaded are tested here.  The cases 
# where the database can be successfully loaded are tested with the loaders 
# themselves.

@parametrize_from_file(
        schema=Schema({
            'config': {str: eval},
            'error': with_freeze.error,
        }),
)
def test_load_db_err(config, error):
    with error:
        load_db(config=config)
